#include <assert.h>
#include <fstream>
#include <iostream>
#include <stdio.h>
#include <string>
#include <sdsl/construct.hpp>
#include <sdsl/construct_lcp.hpp>
#include <sdsl/int_vector.hpp>
#include <sdsl/rank_support_v.hpp>
#include <math.h>


using namespace sdsl;
using namespace std;

template <class T>
class cached_external_array
{
	private:
		std::string		m_filename;
		T*				m_buffer;
		bool			m_need_to_write;
		uint64_t		m_buffersize;
		uint64_t		m_begin;
		uint64_t		m_end;
		bool			m_readforward;
		bool			m_writeforward;
		bool			m_persistent;

		void read_block(const uint64_t idx)
		{
			m_begin = idx;
			std::ifstream file(m_filename.c_str(), std::ios::in|std::ios::binary|std::ios::ate);
			assert(file.good());
			file.seekg(m_begin*sizeof(T));
			assert(file.good());
			file.read(reinterpret_cast<char*>(m_buffer), sizeof(T)*m_buffersize);
			if(file.eof())
			{
				file.clear();
			}
			assert(file.good());
			file.close();
			assert(file.good());
			m_end = m_begin+m_buffersize;
		}

		void write_current_block()
		{
			if(m_need_to_write)
			{
				std::ofstream file(m_filename.c_str(), std::ios::in|std::ios::out|std::ios::binary|std::ios::ate);
				assert(file.good());
				file.seekp(m_begin*sizeof(T));
				assert(file.good());
				file.write(reinterpret_cast<char*>(m_buffer), sizeof(T)*m_buffersize);
				assert(file.good());
				file.close();
				assert(file.good());
				m_need_to_write = false;
			}
		}

	public:
		cached_external_array()
		{
			m_filename = "";
			m_need_to_write = false;
			m_buffersize = 0;
			m_begin = 0;
			m_end = 0;
			m_readforward = true;
			m_writeforward = true;
			m_persistent = false;
			m_buffer = new T[m_buffersize];
		}

		cached_external_array(std::string _filename, const uint64_t _buffersize=1024*1024, bool read_forward=true, bool write_forward=true)
		{
			m_filename = _filename;
			m_need_to_write = false;
			m_buffersize = _buffersize;
			m_begin = 0;
			m_end = 0;
			m_readforward = read_forward;
			m_writeforward = write_forward;
			m_persistent = false;

			// allocate memory for buffer
			m_buffer = new T[m_buffersize];

			// reset or create file
			std::ofstream file(m_filename.c_str(), std::ios::out|std::ios::binary|std::ios::ate);
			file.close();
		}

		~cached_external_array()
		{
			if(!m_persistent){
				remove(m_filename.c_str());
			}
			delete[] m_buffer;
		}

		void init(std::string _filename, const uint64_t _buffersize=1024*1024, bool read_forward=true, bool write_forward=true)
		{
			m_filename = _filename;
			m_need_to_write = false;
			m_buffersize = _buffersize;
			m_begin = 0;
			m_end = 0;
			m_readforward = read_forward;
			m_writeforward = write_forward;
			m_persistent = false;

			// allocate memory for buffer
			m_buffer = new T[m_buffersize];

			// reset or create file
			std::ofstream file(m_filename.c_str(), std::ios::out|std::ios::binary|std::ios::ate);
			file.close();
		}

		uint64_t size()
		{
			std::ifstream file(m_filename.c_str(), std::ios::in|std::ios::binary|std::ios::ate);
			assert(file.good());
			uint64_t bytes = file.tellg();
			assert(file.good());
			file.close();
			return bytes/sizeof(T);
		}

		void buffersize(uint64_t buffersize)
		{
			write_current_block();
			m_begin = 0;
			m_end = 0;
			m_buffersize = buffersize;
			delete[] m_buffer;
			m_buffer = new T[m_buffersize];
		}

		void set_write_direction(bool write_forward)
		{
			m_writeforward = write_forward;
		}

		void set_persistence(bool persistent)
		{
			m_persistent = persistent;
		}

		void write(const uint64_t idx, const uint64_t value)
		{
			// Check if one must first load correct block
			if(idx < m_begin or m_end <= idx)
			{
				// write current block first
				write_current_block();
				if(m_writeforward)
				{
					read_block(idx);
				}
				else
				{
					if(idx < m_buffersize)
					{
						read_block(0);
					}
					else
					{
						read_block(idx-m_buffersize+1);
					}
				}
			}
			m_buffer[idx-m_begin] = value;
			m_need_to_write = true;
		}

		void set_read_direction(bool read_forward)
		{
			m_readforward = read_forward;
		}

		void reset()
		{
			m_need_to_write = false;
			m_begin = 0;
			m_end = 0;

			// reset file
			remove(m_filename.c_str());
			std::ofstream file(m_filename.c_str(), std::ios::out|std::ios::binary|std::ios::ate);
			file.close();
		}

		uint64_t read(const uint64_t idx)
		{
			// Check if one must first load correct block
			if(idx < m_begin or m_end <= idx)
			{
				write_current_block();
				if(m_readforward)
				{
					read_block(idx);
				}
				else
				{
					if(idx < m_buffersize)
					{
						read_block(0);
					}
					else
					{
						read_block(idx-m_buffersize+1);
					}
				}
			}
			return m_buffer[idx-m_begin];
		}
		void flush()
		{
			write_current_block();
		}
		void swap(cached_external_array& cea)
		{
			if (this != &cea) {
				std::swap(m_filename, cea.m_filename);
				std::swap(m_buffer, cea.m_buffer);
				std::swap(m_need_to_write, cea.m_need_to_write);
				std::swap(m_buffersize, cea.m_buffersize);
				std::swap(m_begin, cea.m_begin);
				std::swap(m_end, cea.m_end);
				std::swap(m_readforward, cea.m_readforward);
				std::swap(m_writeforward, cea.m_writeforward);
				std::swap(m_persistent, cea.m_persistent);
			}
		}
};


template<class int_vector_type>
inline uint64_t get_next_lms_position(int_vector_type &text, uint64_t i)
{
	if(i >= text.size()-3)
	{
		return text.size()-1;
	}
	// text[i] is S-TYP or L-TYP
	while(text[i] <= text[i+1])
	{
		++i;
	}
	// text[i] is L-TYP or S-TYP
	uint64_t candidate = i+1;
	while(text[i] >= text[i+1])
	{
		if(text[i] > text[i+1])
		{
			if(i+1==text.size()-1) return text.size()-1;
			candidate = i+1;
		}
		++i;
	}
	return candidate;
}

void construct_sa_simple(int_vector<> &text, int_vector<> &sa, size_t n, size_t text_offset, size_t sigma, uint64_t rekursion)
{
	std::cout << std::string(rekursion, '\t') << "recursion level=" << rekursion << " n=" << n << " text.width=" << (uint64_t)text.width() << " text_offset=" << text_offset << " sigma=" << sigma << std::endl;

	uint64_t buffersize = 1024*1024/8;
	cached_external_array<uint64_t> c_array("c_array_"+util::to_string(rekursion), buffersize, true, true);

	size_t name = 0;
	size_t number_of_lms_strings = 0;
	// Phase 1
	{
		std::vector<uint64_t> bkt(sigma, 0);
		// Step 1 - Count characters into c array
		for(size_t i=0; i<n; ++i)
		{
			++bkt[text[text_offset+i]];
		}

		// Step 1.5 save them into cached_external_array
		for(size_t c=0; c<sigma; ++c)
		{
			c_array.write(c, bkt[c]);
		}

		// Step 2 Calculate End-Pointer of Buckets
		bkt[0] = 0;
		for(size_t c=1; c<sigma; ++c)
		{
			bkt[c] = bkt[c-1]+bkt[c];
		}

		// Step 3 - Insert S*-positions into correct bucket of SA but not in correct order inside the buckets
		for(size_t i=n-2, was_s_typ = 1; i<n; --i)
		{
			if(text[text_offset+i]>text[text_offset+i+1])
			{
				if(was_s_typ)
				{
					sa[bkt[text[text_offset+i+1]]--] = i+1;
					++number_of_lms_strings;
					was_s_typ = 0;
				}
			}
			else if(text[text_offset+i]<text[text_offset+i+1])
			{
				was_s_typ = 1;
			}
		}

		// Step 4 - Calculate Begin-Pointer of Buckets
		bkt[0] = 0;
		for(size_t c=1; c<sigma; ++c)
		{
			bkt[c] = bkt[c-1]+c_array.read(c-1);
		}

		// Step 5 - Scan from Left-To-Right to induce L-Types
		for(size_t i=0; i<n; ++i)
		{
			if(sa[i] > 0 and text[text_offset+ sa[i] ] <= text[text_offset+ sa[i]-1 ]) // schneller als  if(sa[i]>0 and bkt_beg[text[ sa[i]-1 ]] > i)
			{
				sa[bkt[text[text_offset+ sa[i]-1 ]]++] = sa[i]-1;
				sa[i] = 0;
			}
		}

		// Step 6 - Scan from Right-To-Left to induce S-Types
		bkt[0] = 0;
		for(size_t c=1; c<sigma; ++c)
		{
			bkt[c] = bkt[c-1]+c_array.read(c);
		}

		for(size_t i=n-1, endpointer=n; i<n; --i)
		{
			if(sa[i]>0)
			{
				if(text[text_offset+ sa[i]-1 ] <= text[text_offset+ sa[i] ]) // schneller als    if(bkt_end[text[ sa[i]-1 ]] < i)
				{
					sa[bkt[text[text_offset+ sa[i]-1 ]]--] = sa[i]-1;
				}
				else
				{
					sa[--endpointer] = sa[i];
				}
				sa[i] = 0;
			}
		}

		// Step 7 - Determine length of LMS-Strings
		for(size_t i=n-2, end=n-2, was_s_typ = 1; i<n; --i)
		{
			if(text[text_offset+i]>text[text_offset+i+1])
			{
				if(was_s_typ)
				{
					sa[(i+1)>>1] = end-i;
					end = i+1;
					was_s_typ = 0;
				}
			}
			else if(text[text_offset+i]<text[text_offset+i+1])
			{
				was_s_typ = 1;
			}
		}

		// Step 8 - Rename
		for(size_t i=n-number_of_lms_strings+1, cur_pos=0, cur_len=0, last_pos=n-1, last_len=1; i<n; ++i)
		{
			cur_pos = sa[i];
			cur_len = sa[(cur_pos>>1)];
			if(cur_len == last_len)
			{
				size_t l = 0;
				while(l < cur_len and text[text_offset+cur_pos+l] == text[text_offset+last_pos+l])
				{
					++l;
				}
				if(l >= cur_len)
				{
					--name;
				}
			}
			sa[(cur_pos>>1)] = ++name;
			last_pos = cur_pos;
			last_len = cur_len;
		}
	}

	// Step 9 - Calculate SA of new string - in most cases recursive
	if(name+1 < number_of_lms_strings)
	{
		// Move Names to the end
		for(size_t i=0, t=n-number_of_lms_strings; i<(n>>1); ++i)
		{
			if(sa[i] > 0)
			{
				sa[t++] = sa[i];
				sa[i] = 0;
			}
		}
		sa[n-1] = 0;

		// Recursive call
		c_array.buffersize(0);
		construct_sa_simple(sa, sa, number_of_lms_strings, n-number_of_lms_strings, name+1, rekursion+1);
		c_array.buffersize(buffersize);
		for(size_t i=n-2, endpointer = n-1, was_s_typ = 1; i<n; --i)
		{
			if(text[text_offset+i]>text[text_offset+i+1])
			{
				if(was_s_typ)
				{
					sa[endpointer--] = i+1;
					was_s_typ = 0;
				}
			}
			else if(text[text_offset+i]<text[text_offset+i+1])
			{
				was_s_typ = 1;
			}
		}

		// Sort S*-positions in correct order into SA
		for(size_t i=0; i<number_of_lms_strings; ++i)
		{
			size_t pos = sa[i];
			sa[i] = sa[n-number_of_lms_strings+pos];
			sa[n-number_of_lms_strings+pos] = 0;
		}
	}
	else
	{
		// Move s*-Positions to front
		sa[0] = n-1;
		for(size_t i=1; i<number_of_lms_strings; ++i)
		{
			sa[i] = sa[n-number_of_lms_strings+i];
			sa[n-number_of_lms_strings+i] = 0;
		}
		// Clear lex. names
		for(size_t i=number_of_lms_strings; i<(n>>1); ++i)
		{
			sa[i] = 0;
		}
	}

	// Phase 3
	{
		// Step 10 - Count characters into c array

		// Step 11 - Calculate End-Pointer of Buckets
		std::vector<uint64_t> bkt(sigma, 0);
		for(size_t c=1; c<sigma; ++c)
		{
			bkt[c] = bkt[c-1]+c_array.read(c);
		}

		// Step 12 - Move S*-positions in correct order into SA
		for(size_t i=number_of_lms_strings-1; i<n; --i)
		{
			size_t pos = sa[i];
			sa[i] = 0;
			sa[ bkt[text[text_offset+pos]]-- ] = pos;
		}

		// Step 13 - Calculate Begin-Pointer of Buckets
		bkt[0] = 0;
		for(size_t c=1; c<sigma; ++c)
		{
			bkt[c] = bkt[c-1]+c_array.read(c-1);
		}

		// Step 14 - Scan from Left-To-Right to induce L-Types
		for(size_t i=0; i<n; ++i)
		{
			if(sa[i] > 0 and text[text_offset+ sa[i] ] <= text[text_offset+ sa[i]-1 ]) // schneller als  if(sa[i]>0 and bkt_beg[text[ sa[i]-1 ]] > i)
			{
				sa[bkt[text[text_offset+ sa[i]-1 ]]++] = sa[i]-1;
			}
		}

		// Step 15 - Scan from Right-To-Left to induce S-Types
		bkt[0] = 0;
		for(size_t c=1; c<sigma; ++c)
		{
			bkt[c] = bkt[c-1]+c_array.read(c);
		}
		for(size_t i=n-1; i<n; --i)
		{
			if(sa[i] > 0 and text[text_offset+sa[i]-1] <= text[text_offset+sa[i]])
			{
				sa[bkt[text[text_offset+ sa[i]-1 ]]--] = sa[i]-1;
			}
		}
	}
	std::cout << std::string(rekursion, '\t') << "recursion level=" << rekursion << " n=" << n << " text.width=" << (uint64_t)text.width() << " text_offset=" << text_offset << " sigma=" << sigma << std::endl;

	return;
}

template <class T_charsize, class T_nsize, class int_vector_type >
void _construct_bwt_from_text(int_vector_type &text, int_vector_type &bwt, uint64_t sigma, uint64_t rekursion)
{
	std::string text_filename = "text_rek_"+util::to_string(rekursion);
	store_to_file(text, text_filename);
	uint64_t n = text.size();
	uint64_t buffersize = 1024*1024/8;
	uint8_t int_width = bits::hi(n-1)+1;
	std::cout << std::string(rekursion, '\t') << "recursion level=" << rekursion << " n=" << n << " text.width=" << (uint64_t)text.width() << " sigma=" << sigma << std::endl;

	std::cout << std::string(rekursion, '\t') << "Step 1 - Scan Text from right to left and count LMS, S and L characters" << std::endl;

	// define variable
	size_t first_lms_pos=0;
	size_t number_of_lms_strings = 0;
	int_vector<> bkt_lms(sigma, 0, int_width);
	int_vector<> C(sigma, 0, int_width);
	size_t bkt_s_last = 0, bkt_s_sum=0, bound_s=0;
	int_vector<> bkt_s(sigma, 0, int_width);
	size_t bkt_l_sum = 0;
	size_t parts = 0;
	if(rekursion == 0 or rekursion == 1)
	{
 		parts = 10;
	}

	int_vector<> bkt_l(sigma, 0, int_width);
	C[text[n-1]] = C[text[n-1]]+1;
	bool was_s_typ = 1;
	for(size_t i=n-2; i+1>0; --i)
	{
		C[text[i]] = C[text[i]]+1;
		if(was_s_typ)
		{
			bkt_s[text[i+1]] = bkt_s[text[i+1]]+1;
			if(text[i]>text[i+1])
			{
				bkt_lms[text[i+1]] = bkt_lms[text[i+1]] + 1;
				++number_of_lms_strings;
				first_lms_pos = i+1;
				was_s_typ = 0;
			}
		}
		else
		{
			if(text[i]<text[i+1])
			{
				was_s_typ = 1;
			}
		}
	}
	if(was_s_typ)
	{
		bkt_s[text[0]] = bkt_s[text[0]]+1;
	}
	bkt_l[0] = C[0]-bkt_s[0];
	for(size_t i=1; i<C.size(); ++i)
	{
		bkt_l[i] = C[i]-bkt_s[i];
		C[i] = C[i]+C[i-1];
	}

	std::cout << std::string(rekursion, '\t') << "Step 2 - Scan Text from right to left and detect LMS-Positions. Sort and write them to disk" << std::endl;

	cached_external_array<T_nsize> right("right_"+util::to_string(rekursion), buffersize, false, true);
	size_t right_pointer=0;
	cached_external_array<T_nsize> left("left_"+util::to_string(rekursion), buffersize, false, true);
	size_t left_pointer=0;
	{
		for(size_t i=0, tmp2=0, tmp=0; i<sigma; ++i)
		{
			tmp += bkt_lms[i];
			bkt_lms[i] = tmp2;
			tmp2 = tmp;
		}
		cached_external_array<T_nsize> lms_positions("lms_positions_"+util::to_string(rekursion), buffersize, true, true);
		size_t lms_positions_pointer=0;
		for(size_t i=n-2, was_s_typ = 1; i+1>0; --i)
		{
			if(text[i]>text[i+1])
			{
				if(was_s_typ)
				{
					lms_positions.write(lms_positions_pointer++, bkt_lms[text[i+1]]);
					lms_positions.write(lms_positions_pointer++, i+1);
					bkt_lms[text[i+1]] = bkt_lms[text[i+1]] + 1;
					was_s_typ = 0;
				}
			}
			else if(text[i]<text[i+1])
			{
				was_s_typ = 1;
			}
		}
		util::clear(text);
		{
			// Order lms_positions according to first character
			int_vector<> lms_strings(number_of_lms_strings, 0, int_width);
			for(size_t i=0; i<lms_positions_pointer; )
			{
				size_t idx = lms_positions.read(i++);
				size_t val = lms_positions.read(i++);
				lms_strings[idx] = val;
			}
			// Store it to file
			for(left_pointer=0; left_pointer<number_of_lms_strings; ++left_pointer)
			{
				left.write(left_pointer, lms_strings[number_of_lms_strings-left_pointer-1]);
			}
		}
		load_from_file(text, text_filename);
	}
	left_pointer--;


	std::cout << std::string(rekursion, '\t') << "Step 3 - Scan virtual array from left to right" << std::endl;
	{
		// prepare bkt_l and backup it into bkt_lms
		for(size_t i=0, tmp=0; i<sigma; ++i)
		{
			tmp = bkt_l[i];
			bkt_l[i] = bkt_l_sum;
			bkt_l_sum += tmp;
			bkt_lms[i] = bkt_l[i];
		}

		size_t partsize = bkt_l_sum/parts+1;

		int_vector<> array(partsize, 0, int_width);
		std::vector< cached_external_array<T_nsize> > cached_array(parts-1);
		std::vector<uint64_t> cached_array_pointer(cached_array.size());
		for(size_t i=1; i<parts; ++i)
		{
			cached_array[i-1].init("rightbuffer_"+ util::to_string(i-1) +"_"+util::to_string(rekursion), buffersize, true, true);
			cached_array_pointer[i-1] = 0;
		}

		for(size_t c=0, pos=0, offset=0; c<sigma; ++c)
		{
			// begin with array
			for( ; pos<bkt_l[c]; ++pos)
			{
				// Load lazy values
				if(pos-offset >= partsize)
				{
					offset += partsize;
					for(size_t i=0, cur_part=pos/partsize-1; i<cached_array_pointer[cur_part]/2; ++i)
					{
						size_t src = cached_array[cur_part].read(2*i);
						size_t val = cached_array[cur_part].read(2*i+1);
						array[src-offset] = val;
					}
					cached_array[pos/partsize-1].reset();
				}

				size_t idx = array[pos-offset];
				if(idx == 0)
				{
					right.write(right_pointer++, idx);
				}
				else
				{
					size_t symbol = text[idx-1];
					if(symbol >= c)
					{
						size_t val = idx-1;
						size_t src = bkt_l[symbol];
						bkt_l[symbol] = bkt_l[symbol] + 1;
						if( (src-offset)/partsize == 0)
						{
							array[src-offset] = val;
						}
						else
						{
							size_t part = src/partsize-1;
							cached_array[part].write(cached_array_pointer[part]++, src);
							cached_array[part].write(cached_array_pointer[part]++, val);
						}
					}
					else
					{
						right.write(right_pointer++, idx);
					}
				}
			}
			// continue with stack
			while(left_pointer < number_of_lms_strings and text[left.read(left_pointer)] == c)
			{
				size_t idx = left.read(left_pointer--);
				if(idx == 0)
				{
					idx = n;
				}
				--idx;
				size_t symbol = text[idx];

				size_t val = idx;
				size_t src = bkt_l[symbol];
				bkt_l[symbol] = bkt_l[symbol] + 1;
				if( (src-offset)/partsize == 0)
				{
					array[src-offset] = val;
				}
				else
				{
					size_t part = src/partsize-1;
					cached_array[part].write(cached_array_pointer[part]++, src);
					cached_array[part].write(cached_array_pointer[part]++, val);
				}
			}
		}

		// Restore bkt_l from bkt_lms
		for(size_t i=0; i<sigma; ++i)
		{
			bkt_l[i] = bkt_lms[i];
		}
	}
	right_pointer--;


	std::cout << std::string(rekursion, '\t') << "Step 4 - Scan virtual array from right to left" << std::endl;
	left_pointer = 0;
	left.reset();
	{
		// Prepare bkt_s and backup it into bkt_lms
		bkt_s_last=0, bkt_s_sum=0;
		for(size_t i=0; i<sigma; ++i)
		{
			bkt_s_sum += bkt_s[i];
			if(bkt_s[i])
			{
				bkt_s[i] = bkt_s_sum;
				bkt_s_last = bkt_s_sum;
			}
			else
			{
				bkt_s[i] = bkt_s_sum;
			}
			bkt_lms[i] = bkt_s[i];
		}
		bound_s = bkt_s_sum;

		// determine splitting parameters
		for(size_t i=0; i<sigma; ++i)
		{
			if(bkt_s[i] > bkt_s_sum/2)
			{
				bkt_s_sum = bkt_s[i];
				break;
			}
		}

		size_t partsize = bound_s/parts+1;
		int_vector<> array(partsize, 0, int_width);
		std::vector< cached_external_array<T_nsize> > cached_array(parts-1);
		std::vector<uint64_t> cached_array_pointer(cached_array.size());
		for(size_t i=0; i<parts-1; ++i)
		{
			cached_array[i].init("leftbuffer_"+ util::to_string(i) +"_"+util::to_string(rekursion), buffersize, true, true);
			cached_array_pointer[i] = 0;
		}

		for(size_t c=sigma-1, pos=bkt_s_last-1, offset=partsize*(parts-1); c<sigma; --c)
		{
			// begin with array
			for(; pos+1 > bkt_s[c]; --pos)
			{
				while(pos < offset)
				{
					// Load buffered values
					offset -= partsize;
					// offset = pos-partsize;
					for(size_t i=0, cur_part=offset/partsize; i<cached_array_pointer[cur_part]/2; ++i)
					{
						size_t src = cached_array[cur_part].read(2*i);
						size_t val = cached_array[cur_part].read(2*i+1);
						array[src-offset] = val;
					}
					cached_array[offset/partsize].reset();
				}

				size_t idx = array[pos-offset];
				if(idx==0)
				{
					idx = n;
				}
				--idx;
				size_t symbol = text[idx];
				if(symbol <= c)
				{
					bkt_s[symbol] = bkt_s[symbol] - 1;
					size_t val = idx;
					size_t src = bkt_s[symbol];
					if( src >= offset)
					{
						array[src-offset] = val;
					}
					else
					{
						size_t part = src/partsize;
						cached_array[part].write(cached_array_pointer[part]++, src);
						cached_array[part].write(cached_array_pointer[part]++, val);
					}
				}
				else
				{
					left.write(left_pointer++, array[pos-offset]);
				}
			}

			// continue with stack
			while(right_pointer < number_of_lms_strings and text[right.read(right_pointer)] == c)
			{
				size_t idx = right.read(right_pointer--);
				if(idx == 0)
				{
					idx = n;
				}
				--idx;
				size_t symbol = text[idx];
				bkt_s[symbol] = bkt_s[symbol] - 1;

				size_t val = idx;
				size_t src = bkt_s[symbol];
				if( src >= offset)
				{
					array[src-offset] = val;
				}
				else
				{
					size_t part = src/partsize;
					cached_array[part].write(cached_array_pointer[part]++, src);
					cached_array[part].write(cached_array_pointer[part]++, val);
				}
			}
		}

		// Restore bkt_s from bkt_lms
		for(size_t i=0; i<sigma; ++i)
		{
			bkt_s[i] = bkt_lms[i];
		}
	}
	right.reset();
	right_pointer = 0;
	--left_pointer;


	std::cout << std::string(rekursion, '\t') << "Step 5 - Detect same lms-Strings, write text to file" << std::endl;

	int_vector<1> same_lms(number_of_lms_strings, false);
	size_t last_end_pos = first_lms_pos, ordnung = number_of_lms_strings-1;
	same_lms[number_of_lms_strings-1] = true;
	for(size_t i=number_of_lms_strings-2, a=0, b=0, last_a=left.read(number_of_lms_strings-1); i<number_of_lms_strings; --i)
	{
		b = last_a;
		a = left.read(i);
		last_a = a;

		size_t end_pos = get_next_lms_position(text, a);
		if(end_pos-a == last_end_pos-b)
		{
			while(a < end_pos and text[a] == text[b])
			{
				++a;
				++b;
			}
			if(text[a] == text[b])
			{
				same_lms[i] = true;
				--ordnung;
			}
		}
		last_end_pos = end_pos;
	}


	std::cout << std::string(rekursion, '\t') << "Step 6 - Create renaming table and store to file" << std::endl;

	std::string filename_translation = "translation_"+util::to_string(rekursion);
	{
		int_vector<> name_to_position(ordnung+1, 0, int_width);
		ordnung = 0;
		name_to_position[ordnung] = first_lms_pos;
		for(size_t i=number_of_lms_strings-1; i<number_of_lms_strings; --i)
		{
			if(!same_lms[i])
			{
				++ordnung;
				name_to_position[ordnung] = get_next_lms_position(text, left.read(i));
			}
		}
		store_to_file(name_to_position, filename_translation);
	}
	util::clear(text);


	std::cout << std::string(rekursion, '\t') << "Step 7 - Create renamed string" << std::endl;

	int_vector<> text_rek;
	if(rekursion==0)
	{
		text_rek.width((bits::hi(ordnung+1)+1));
	}
	else
	{
		text_rek.width((bits::hi(number_of_lms_strings+1)+1));
	}
	text_rek.resize(number_of_lms_strings);
	util::_set_zero_bits(text_rek);
	{
		if(rekursion==0 and n/2*text_rek.width()>8*n)
		{
			size_t size_of_part = n/4+3;
			text_rek.resize(size_of_part);
			util::_set_zero_bits(text_rek);
			ordnung = 0;
			for(size_t i=number_of_lms_strings-1; i<number_of_lms_strings; --i)
			{
				if(!same_lms[i])
				{
					++ordnung;
				}
				if(left.read(i)/2 >= size_of_part)
				{
					text_rek[(left.read(i)/2)-size_of_part] = ordnung;
				}
			}
			std::string filename_text_rek_part2 = "text_rek_part2_"+util::to_string(rekursion);
			size_t pos=0;
			for(size_t i=0; i<size_of_part; ++i)
			{
				if(text_rek[i]>0)
				{
					text_rek[pos++] = text_rek[i];
				}
			}
			text_rek.resize(pos);
			store_to_file(text_rek, filename_text_rek_part2);
			text_rek.resize(size_of_part);
			util::_set_zero_bits(text_rek);
			ordnung = 0;
			for(size_t i=number_of_lms_strings-1; i<number_of_lms_strings; --i)
			{
				if(!same_lms[i])
				{
					++ordnung;
				}
				if(left.read(i)/2 < size_of_part)
				{
					text_rek[left.read(i)/2] = ordnung;
				}
			}
			pos = 0;
			for(size_t i=0; i<size_of_part; ++i)
			{
				if(text_rek[i]>0)
				{
					text_rek[pos++] = text_rek[i];
				}
			}
			text_rek.resize(number_of_lms_strings);
			int_vector_file_buffer<> buf(filename_text_rek_part2, 1024*1024);
			buf.reset();
			size_t part2_size = buf.int_vector_size;
			for(size_t i=0, r_sum=0, r=0; r_sum < part2_size; )
			{
				for( ; i < r_sum+r ; ++i)
				{
					text_rek[pos++] = buf[i-r_sum];
				}
				r_sum += r;
				r = buf.load_next_block();
			}
			text_rek[number_of_lms_strings-1] = 0;
			remove(filename_text_rek_part2.c_str());
		}
		else
		{
			text_rek.resize(n/2+1);
			util::_set_zero_bits(text_rek);
			ordnung = 0;
			for(size_t i=number_of_lms_strings-1; i<number_of_lms_strings; --i)
			{
				if(!same_lms[i])
				{
					++ordnung;
				}
				text_rek[left.read(left_pointer--)/2] = ordnung;
			}
			for(size_t i=0, pos=0; i<text_rek.size(); ++i)
			{
				if(text_rek[i]>0)
				{
					text_rek[pos++] = text_rek[i];
				}
			}
			text_rek[number_of_lms_strings-1] = 0;
			text_rek.resize(number_of_lms_strings);
		}
	}
	util::clear(same_lms);
	left.reset();


	std::cout << std::string(rekursion, '\t') << "Step 8 - Determine complete LMS order (recursivly)" << std::endl;

	int_vector<> bwt_rek;
	bwt_rek.width(text_rek.width());
	if(text_rek.size() > ordnung+1)
	{
		if(rekursion==0)
		{
			if(ordnung+1 < (1ULL<<8))
			{
				if(text_rek.size() < (1ULL<<32))
				{
					_construct_bwt_from_text<uint8_t, uint32_t, int_vector<> >(text_rek, bwt_rek, ordnung+1, rekursion+1);
				}
				else
				{
					_construct_bwt_from_text<uint8_t, uint64_t, int_vector<> >(text_rek, bwt_rek, ordnung+1, rekursion+1);
				}
			}
			else if(ordnung+1 < (1ULL<<16))
			{
				if(text_rek.size() < (1ULL<<32))
				{
					_construct_bwt_from_text<uint16_t, uint32_t, int_vector<> >(text_rek, bwt_rek, ordnung+1, rekursion+1);
				}
				else
				{
					_construct_bwt_from_text<uint16_t, uint64_t, int_vector<> >(text_rek, bwt_rek, ordnung+1, rekursion+1);
				}
			}
			else if(ordnung+1 < (1ULL<<32))
			{
				if(text_rek.size() < (1ULL<<32))
				{
					_construct_bwt_from_text<uint32_t, uint32_t, int_vector<> >(text_rek, bwt_rek, ordnung+1, rekursion+1);
				}
				else
				{
					_construct_bwt_from_text<uint32_t, uint64_t, int_vector<> >(text_rek, bwt_rek, ordnung+1, rekursion+1);
				}
			}
			else
			{
				_construct_bwt_from_text<uint64_t, uint64_t, int_vector<> >(text_rek, bwt_rek, ordnung+1, rekursion+1);
			}
		}
		else
		{
			text_rek.resize(text_rek.size()*2);
			for(size_t i=0; i<number_of_lms_strings; ++i)
			{
				text_rek[number_of_lms_strings+i] = text_rek[i];
				text_rek[i] = 0;
			}
			construct_sa_simple(text_rek, text_rek, number_of_lms_strings, number_of_lms_strings, ordnung+1, rekursion+1);
			for(size_t i=0; i<number_of_lms_strings; ++i)
			{
				if(text_rek[i]>0)
				{
					text_rek[i] = text_rek[number_of_lms_strings+text_rek[i]-1];
				}
				else
				{
					text_rek[i] = 0;
				}
			}
			text_rek.resize(number_of_lms_strings);
			bwt_rek.resize(number_of_lms_strings);
			for(size_t i=0; i<number_of_lms_strings; ++i)
			{
				bwt_rek[i] = text_rek[i];
			}
		}
	}
	else
	{
		bwt_rek.resize(text_rek.size());
		bwt_rek[text_rek[0]] = text_rek[text_rek.size()-1];
		for(size_t i=1; i<text_rek.size(); ++i)
		{
			bwt_rek[text_rek[i]] = text_rek[i-1];
		}
	}
	util::clear(text_rek);

	std::cout << std::string(rekursion, '\t') << "Step 9 - Retranslate bwt" << std::endl;

	{
		left_pointer = 0;
		int_vector<> name_to_position;
		load_from_file(name_to_position, filename_translation);
		remove(filename_translation.c_str());
		for(size_t i=bwt_rek.size()-1; i+1>0; --i)
		{
			left.write(left_pointer++, name_to_position[bwt_rek[i]]);
		}
		left_pointer--;
	}

	std::cout << std::string(rekursion, '\t') << "Step 10 - Reload text" << std::endl;

	util::clear(bwt_rek);
	load_from_file(text, text_filename);
	remove(text_filename.c_str());

	std::cout << std::string(rekursion, '\t') << "Step 12 - Scan virtual array from left to right second time" << std::endl;

	right.reset();
	right_pointer = 0;
	cached_external_array<T_charsize> charstack("charstack_"+util::to_string(rekursion), buffersize, false, true);
	size_t charstack_pointer = 0;
	cached_external_array<T_charsize> cached_bwt("bwt_"+util::to_string(rekursion), buffersize, false, true);
	size_t bwt_pointer = 0;
	{
		size_t partsize = bkt_l_sum/parts+1;

		int_vector<> array(partsize, 0, int_width);
		std::vector< cached_external_array<T_nsize> > cached_array(parts-1);
		std::vector<uint64_t> cached_array_pointer(cached_array.size());
		for(size_t i=1; i<parts; ++i)
		{
			cached_array[i-1].init("rightbuffer_"+ util::to_string(i-1) +"_"+util::to_string(rekursion), buffersize, true, true);
			cached_array_pointer[i-1] = 0;
		}

		for(size_t c=0, pos=0, offset=0; c<sigma; ++c)
		{
			// begin with array
			for( ; pos<bkt_l[c]; ++pos)
			{
				// Load lazy values
				if(pos-offset >= partsize)
				{
					offset += partsize;
					for(size_t i=0, cur_part=pos/partsize-1; i<cached_array_pointer[cur_part]/2; ++i)
					{
						size_t src = cached_array[cur_part].read(2*i);
						size_t val = cached_array[cur_part].read(2*i+1);
						array[src-offset] = val;
					}
					cached_array[pos/partsize-1].reset();
				}

				size_t idx = array[pos-offset];
				if(idx == 0)
				{
					cached_bwt.write(bwt_pointer++, 0);
					right.write(right_pointer++, idx);
				}
				else
				{
					size_t symbol = text[idx-1];
					cached_bwt.write(bwt_pointer++, symbol);
					if(symbol >= c)
					{
						size_t val = idx-1;
						size_t src = bkt_l[symbol];
						bkt_l[symbol] = bkt_l[symbol] + 1;
						if( (src-offset)/partsize == 0)
						{
							array[src-offset] = val;
						}
						else
						{
							size_t part = src/partsize-1;
							cached_array[part].write(cached_array_pointer[part]++, src);
							cached_array[part].write(cached_array_pointer[part]++, val);
						}
					}
					else
					{
						right.write(right_pointer++, idx);
					}
				}
			}
			bwt_pointer = C[c];

			// continue with stack
			while(left_pointer < number_of_lms_strings and text[left.read(left_pointer)] == c)
			{
				size_t idx = left.read(left_pointer--);
				if(idx == 0)
				{
					idx = n;
				}
				--idx;
				size_t symbol = text[idx];
				charstack.write(charstack_pointer++, symbol);

				size_t val = idx;
				size_t src = bkt_l[symbol];
				bkt_l[symbol] = bkt_l[symbol] + 1;
				if( (src-offset)/partsize == 0)
				{
					array[src-offset] = val;
				}
				else
				{
					size_t part = src/partsize-1;
					cached_array[part].write(cached_array_pointer[part]++, src);
					cached_array[part].write(cached_array_pointer[part]++, val);
				}
			}
		}
	}
	left.reset();
	charstack_pointer--;
	right_pointer--;


	std::cout << std::string(rekursion, '\t') << "Step 13 - Scan virtual array from right to left second time " << std::endl;

	cached_bwt.set_write_direction(false);
	{
		size_t partsize = bound_s/parts+1;

		int_vector<> array(partsize, 0, int_width);
		std::vector< cached_external_array<T_nsize> > cached_array(parts-1);
		std::vector<uint64_t> cached_array_pointer(cached_array.size());
		for(size_t i=0; i<parts-1; ++i)
		{
			cached_array[i].init("leftbuffer_"+ util::to_string(i) +"_"+util::to_string(rekursion), buffersize, true, true);
			cached_array_pointer[i] = 0;
		}

		for(size_t c=sigma-1, pos=bkt_s_last-1, offset=partsize*(parts-1); c<sigma; --c)
		{
			// begin with array
			assert(c < C.size() );
			bwt_pointer = C[c]-1;
			for(; pos+1 > bkt_s[c]; --pos)
			{
				while(pos < offset)
				{
					// Load buffered values
					offset -= partsize;
					for(size_t i=0, cur_part=offset/partsize; i<cached_array_pointer[cur_part]/2; ++i)
					{
						size_t src = cached_array[cur_part].read(2*i);
						size_t val = cached_array[cur_part].read(2*i+1);
						assert( (src-offset) < array.size() );
						array[src-offset] = val;
					}
					assert( (offset/partsize) < cached_array.size() );
					cached_array[offset/partsize].reset();
				}

				assert( (pos-offset) < array.size() );
				size_t idx = array[pos-offset];
				if(idx==0)
				{
					idx = n;
				}
				--idx;
				assert( (idx) < text.size() );
				size_t symbol = text[idx];
				if(symbol <= c)
				{
					cached_bwt.write(bwt_pointer--, symbol);
					assert( (symbol) < bkt_s.size() );
					bkt_s[symbol] = bkt_s[symbol] - 1;
					size_t val = idx;
					size_t src = bkt_s[symbol];
					if( src >= offset)
					{
						assert( (src-offset) < array.size() );
						array[src-offset] = val;
					}
					else
					{
						size_t part = src/partsize;
						assert( part < cached_array.size() );
						cached_array[part].write(cached_array_pointer[part]++, src);
						cached_array[part].write(cached_array_pointer[part]++, val);
					}
				}
				else
				{
					cached_bwt.write(bwt_pointer--, charstack.read(charstack_pointer--));
				}
			}

			// continue with stack
			while(right_pointer < number_of_lms_strings and text[right.read(right_pointer)] == c)
			{
				size_t idx = right.read(right_pointer--);
				if(idx == 0)
				{
					idx = n;
				}
				--idx;
				size_t symbol = text[idx];
				assert( (symbol) < bkt_s.size() );
				bkt_s[symbol] = bkt_s[symbol] - 1;

				size_t val = idx;
				size_t src = bkt_s[symbol];
				if( src >= offset)
				{
					assert( (src-offset) < array.size() );
					array[src-offset] = val;
				}
				else
				{
					size_t part = src/partsize;
					assert( (part) < cached_array.size() );
					cached_array[part].write(cached_array_pointer[part]++, src);
					cached_array[part].write(cached_array_pointer[part]++, val);
				}
			}
		}
	}


	std::cout << std::string(rekursion, '\t') << "Step 14 - Load bwt from file to memory" << std::endl;

	cached_bwt.set_read_direction(true);
	text.swap(bwt);
	for(size_t i=0; i<n; ++i)
	{
		bwt[i] = cached_bwt.read(i);
	}

	std::cout << std::string(rekursion, '\t') << "recursion level=" << rekursion << " n=" << n << " text.width=" << (uint64_t)text.width() << " sigma=" << sigma << std::endl;

	return;
}

struct triple
{
	size_t value;
	size_t pos;
	size_t pos_ex;
	size_t nr;
	triple(size_t _value, size_t _pos, size_t _pos_end, size_t _nr)
	{
		value = _value;
		pos = _pos;
		pos_ex = _pos_end;
		nr = _nr;
	}
	triple()
	{
		value = 0;
		pos = 0;
		pos_ex = 0;
		nr = 0;
	}
};

struct text_longs_comp {
	unsigned char * text_l;
	unsigned long long size;
	text_longs_comp(unsigned char * datapointer, unsigned long long text_l_size)
	{
		text_l = datapointer;
		size = text_l_size;
	}
	void show()
	{
		cout << "text_longs:" << endl;
		for(size_t i=0; i<size; ++i)
		{
			cout << i << " " ;
		}
		cout << endl;
		for(size_t i=0; i<size; ++i)
		{
			cout << (size_t)text_l[i] << " " ;
		}
		cout << endl;
	}
	bool equal(triple& a, triple& b)
	{
		if(a.value == b.value)
		{
			if(a.pos>=a.pos_ex and b.pos>=b.pos_ex)
			{
				return true;
			}
			else if(a.pos<a.pos_ex and b.pos<b.pos_ex)
			{
				size_t offset = 0;
				while(text_l[a.pos + offset] == text_l[b.pos + offset] and offset+a.pos < a.pos_ex and offset+b.pos < b.pos_ex)
				{
					++offset;
				}
				if(offset+b.pos==b.pos_ex and offset+a.pos==a.pos_ex)
				{
					return true;
				}
			}
		}
		return false;
	}
	bool operator() (const triple& a, const triple& b)
	{
		if(a.value == b.value)
		{
			if(a.pos>=a.pos_ex)
			{
				return false;
			}
			if(b.pos>=b.pos_ex)
			{
				return true;
			}
			size_t offset = 0;
			while(text_l[a.pos + offset] == text_l[b.pos + offset] and offset+a.pos < a.pos_ex and offset+b.pos < b.pos_ex)
			{
				++offset;
			}
			if(offset+b.pos==b.pos_ex)
			{
				return offset+a.pos<a.pos_ex;
			}
			if(offset+a.pos==a.pos_ex)
			{
				return false;
			}
			return text_l[a.pos + offset] < text_l[b.pos + offset];
		}
		return a.value < b.value;
	}
};

class array_sequential_write
{
	private:
		uint64_t     block_size;
		char *       block;
		uint64_t     m_size;
		uint8_t      value;
		string       filename;
		ofstream     o_file;
		ifstream     i_file;
		bool         forward;
		uint64_t     filepointer; //points to beginning of current block in file
		bool write_block(){
			if(filepointer<0 or filepointer>=m_size){
				return false;
			}
			if(forward){
				if(!o_file.is_open()){
					o_file.open(filename.c_str(), ios::in|ios::binary|ios::ate);
				}
				o_file.seekp(filepointer);
				if(filepointer+block_size<m_size){
					o_file.write(block,block_size);
					clear_block();
					filepointer += block_size;
				}else{
					o_file.write(block,m_size-filepointer);
					filepointer = m_size;
				}
			}else{
				o_file.open(filename.c_str(), ios::in|ios::binary|ios::ate);
				o_file.seekp(filepointer);
				if(filepointer+block_size<m_size){
					o_file.write(block,block_size);
				}else{
					o_file.write(block,m_size-filepointer);
				}
				filepointer = filepointer-block_size;
				o_file.close();
				read_block();
			}
			return true;
		}
		void clear_block(){
			for(unsigned long long i=0; i<block_size; ++i){
				block[i] = value;
			}
		}
		void read_block(){
			if(filepointer+block_size==0 or filepointer==m_size){
				clear_block();
			}else{
				i_file.open(filename.c_str(), ios::in|ios::binary|ios::ate);
				i_file.seekg(filepointer);
				if(filepointer+block_size<m_size){
					i_file.read(block,block_size);
				}else{
					i_file.read(block,m_size-filepointer);
				}
				i_file.close();
			}
			return;
		}
	public:
		array_sequential_write(string _filename, const uint64_t n=1, const uint8_t n_value=0){
			m_size = n;
			value = n_value;
			block_size = 1024*1024;
			block = new char [block_size];
			clear_block();
			filename = _filename;
			start_forward_write();
		}
		~array_sequential_write(){
			if(forward){
			while(filepointer!=m_size){
				write_block();
			}
			}else{
				write_block();
			}
			delete[] block;
			if(o_file.is_open()) o_file.close();
			if(i_file.is_open()) i_file.close();
		}
		void start_forward_write(){  //overwrite file content if any
			o_file.open(filename.c_str(), ios::out|ios::binary);
			forward = true;
			filepointer = 0;
		}
		void start_backward_read_and_write(){  //keep file content, change only the positions set in backward mode
			if(o_file.is_open())o_file.close();
			forward = false;
			filepointer = (m_size/block_size);
			filepointer = filepointer*block_size;
			read_block();
		}
		bool empty() const {
			return m_size == 0;
		}
		bool set(const uint64_t pos, const uint8_t value){
			if(pos>=filepointer and pos<filepointer+block_size){
				block[pos-filepointer] = value;
				return true;
			}
			if(forward and pos>=filepointer+block_size and pos<m_size){
				write_block();
				while(pos>=filepointer+block_size){
					write_block();
				}
				block[pos-filepointer] = value;
				return true;
			}
			if(!forward and pos<filepointer and pos+1>0){
				write_block();
				while(pos<filepointer){
					write_block();
				}
				block[pos-filepointer] = value;
				return true;
			}
			return false;
		}
		uint64_t size() const {
			return m_size;
		}
};

class array_nearly_sequential_read
{
	private:
		uint64_t     block_size;
		unsigned char *       block1;
		unsigned char *       block2;
		uint64_t     m_size;
		uint8_t      value;
		string       filename;
		ofstream     o_file;
		ifstream     i_file;
		bool         forward;
		bool         intvector;
		uint64_t     filepointer; //points to beginning of current block in file
		void read_block(){
			if(forward){
				filepointer += block_size;
				if(filepointer>=m_size){
					clear_block(1);
					clear_block(2);
				}else{
					i_file.seekg(filepointer+block_size);
					assert(i_file.good());
					unsigned char * tmp = block1;
					block1 = block2;
					block2 = tmp;
					if(filepointer+block_size>=m_size){
						clear_block(2);
					}else if(filepointer+2*block_size<m_size){
						i_file.read((char*)block2,block_size);
						assert(i_file.good());
					}else{
						clear_block(2);
						i_file.read((char*)block2,m_size-filepointer);
						assert(i_file.good());
					}
				}
			}else{
				assert(i_file.good());
				filepointer -= block_size;
				if(filepointer+block_size==0){
					clear_block(1);
				}else{
					i_file.close();
					i_file.open(filename.c_str(), ios::in|ios::binary);
					assert(i_file.good());
					i_file.seekg(filepointer);
					assert(i_file.good());
					unsigned char * tmp = block2;
					block2 = block1;
					block1 = tmp;
					if(filepointer+block_size<m_size){
						clear_block(1);
						i_file.read((char*)block1,block_size);
						assert(i_file.good());
					}else{
						clear_block(1);
						i_file.read((char*)block1,m_size-filepointer);
						assert(i_file.good());
					}
				}
			}
			return;
		}
		void clear_block(unsigned char blockNr){
			if(blockNr==1){
				for(unsigned long long i=0; i<block_size; ++i){
					block1[i] = value;
				}
			}else if(blockNr==2){
				for(unsigned long long i=0; i<block_size; ++i){
					block2[i] = value;
				}
			}
		}
		void begin_read(){
			if(m_size<block_size){
				filepointer = 0;
				i_file.seekg (filepointer);
				i_file.read((char*)block1,m_size);
			}else if(m_size<2*block_size){
				filepointer = 0;
				i_file.seekg (filepointer);
				i_file.read((char*)block1,block_size);
				i_file.seekg (filepointer+block_size);
				i_file.read((char*)block2,m_size-filepointer);
			}
			else if(forward){
				filepointer = 0;
				i_file.seekg (0, ios::beg);
				i_file.read((char*)block1,block_size);
				i_file.seekg (block_size);
				i_file.read((char*)block2,block_size);
			}else{
				filepointer = m_size/block_size;
				if(filepointer*block_size==m_size) --filepointer;
				filepointer = filepointer*block_size;
				i_file.seekg (filepointer);
				i_file.read((char*)block2,m_size-filepointer);
				filepointer -= block_size;
				i_file.seekg (filepointer);
				i_file.read((char*)block1,block_size);
			}
		}
	public:
		array_nearly_sequential_read(string _filename, uint64_t _size, bool _forward, uint64_t _block_size, bool _intvector){
			block_size = _block_size;
			block1 = new unsigned char [block_size];
			block2 = new unsigned char [block_size];
			filename = _filename;
			i_file.open(_filename.c_str(), ios::in|ios::binary);
			assert(i_file.good());
			m_size = _size;
			forward = _forward;
			intvector = _intvector;
			if(_intvector) m_size += 8;
			begin_read();
		}
		~array_nearly_sequential_read(){
			if(i_file.is_open()) i_file.close();
			remove(filename.c_str());
			delete[] block1;
			delete[] block2;
		}
		void finish_this_read() {
			i_file.close();
		}
		void begin_backward_read() {
			i_file.open(filename.c_str(), ios::in|ios::binary);
			forward = false;
			begin_read();
		}
		void begin_forward_read() {
			i_file.open(filename.c_str(), ios::in|ios::binary);
			forward = true;
			begin_read();
		}
		bool empty() const {
			return m_size == 0;
		}
		uint64_t size() const {
			return m_size;
		}
		unsigned char at(const uint64_t pos){
			uint64_t i = pos;
			if(intvector) i += 8;
			if(i>=m_size){
				return 0;
			}
			if(i>=filepointer and i<filepointer+block_size){
				return block1[i-filepointer];
			}
			if(i>=filepointer+block_size and i<filepointer+2*block_size){
				return block2[i-filepointer-block_size];
			}
			if(i>=filepointer+2*block_size){
				forward = true;
				read_block();
				while(i>=filepointer+2*block_size){
					read_block();
				}
				return block2[i-filepointer-block_size];
			}
			if(i<filepointer and i+1>0){
				forward = false;
				read_block();
				while(i<filepointer){
					read_block();
				}
				return block1[i-filepointer];
			}
			cerr << "wrong read" << endl;
			return 1;
		}
};



void construct_bwt_smaller_alphabet(string bwt_filename, int_vector<8> &text, vector<bool> &used_chars, size_t lms_short_bound, size_t lms_long_bound)
{
	size_t n = text.size();
	size_t sigma = 0;
	vector<unsigned char> number_to_char(256, 0);
	string text_trans_filename = "text.trans.sdsl";
	{
		vector<int> char_to_number(256, 0);

		// Rename characters
		for(size_t i=0; i<used_chars.size(); ++i)
		{
			if(used_chars[i])
			{
				char_to_number[i] = sigma;
				number_to_char[sigma] = i;
				++sigma;
			}
		}
		--sigma;

		// Transformate Text
		for(size_t i=0; i<text.size(); ++i)
		{
			text[i] = char_to_number[text[i]];
		}
		char* text2 = (char*)text.data();
		ofstream transfile (text_trans_filename.c_str(),ofstream::binary);
		transfile.write(text2,n);
		transfile.close();
	}
	text.resize(0);
	util::clear(text);
	used_chars.resize(0);
	util::clear(used_chars);


	// calculate bwt
	cout << "recursion level=0" << "  n=" << n << "  sigma=" << sigma << "  lms_short_bound=" << lms_short_bound << "  lms_long_bound=" << lms_long_bound << endl;

	// Step 0: Preparation
	cout << "Step 0  - prepare" << endl;

	unsigned long long realsigma = sigma+1; // realsigma is size of alphabet incl $
	if(sigma<=1)
	{
		sigma = 2;
	}
	size_t dollar_pos = 0;
	size_t sigma_rek = 0;                                        // size of alphabet in recursion
	size_t red_text_size = 0;
	size_t name_to_pos_size = 0;
	string filename_translation = "translation_0";
	string filename_redtext = "redtext_0";
	size_t lastlms_name = 0;
	int_vector<> text_rek;                                       // text for recursion
	size_t nrs = 1;                                              // number of lms-strings (there is always the lms-string $...)
	vector<unsigned long long> C(realsigma+1, 0);
	{
		vector<triple> lms_long;                                 // vector to hold one triple for every long lms-string
		bit_vector lms_short(pow(sigma,lms_short_bound), false); // lms_short[i] will be true if a short lms-string with this lms_value exists
		uint64_t block_s = n/200;
		if(block_s<1000) block_s = 1000;
		array_nearly_sequential_read seqtext (text_trans_filename, n, false, block_s, false); // buffer to read translated text
		vector<unsigned char> text_longs;                        // will hold parts of the long lms-strings (those symbols of the long lms-string that are not included in the value) (needed for comparison during sorting)

		// Step 1:
		std::cout << "Step 1a - go through text backwards to find short and long lms strings" << std::endl;

		unsigned long long lms_value;  // value of an lms-string (see explanation "number" in paper "Space-efficient Construction of the Burrows-Wheeler Transform"
		size_t last = n-1;             // position of the lms-string that was found last; initially n-1 because thats the position of $
		bool smaller = false;          // is true if last viewed position was a S-position and the next L will turn it into an S*
		size_t nrs_s = 2;
		size_t longestlong = lms_short_bound;
		size_t startpos = n-2;

		// $
		++C[seqtext.at(n-1)];
		// find last lms string (...$)
		for(size_t i=startpos; i+1>0; --i)
		{
			++C[seqtext.at(i)];
			if(smaller)
			{
				if(seqtext.at(i)>seqtext.at(i+1))
				{
					lms_value = 0;
					for(size_t j=i+1; j<last and j<i+lms_long_bound+1; ++j)
					{
						lms_value = lms_value*sigma+(seqtext.at(j)-1);
					}
					for(size_t j=i+lms_long_bound+1; j<last; ++j)
					{
						text_longs.push_back(seqtext.at(j));
					}
					text_longs.push_back(seqtext.at(last));
					if(last-i-1<lms_long_bound)
					{
						for(size_t j=0; j<lms_long_bound-(last-i-1); ++j)
						{
							lms_value = lms_value*sigma;
						}
					}
					lms_long.push_back(triple(lms_value,0,text_longs.size(),nrs));
					nrs++;
					smaller = false;
					last = i+1;
					startpos = i-1;
					break;
				}
			}
			else
			{
				if(seqtext.at(i)<seqtext.at(i+1))
				{
					smaller = true;
				}
			}
		}

		if(nrs==1) //if the input consists of only one lms_string, no recursion necessary
		{
			// calculate C-Array
			for(unsigned long long i=1, tmp=0, tmp2=0; i<C.size(); ++i)
			{
				tmp = C[i];
				C[i] = tmp2+C[i-1];
				tmp2 = tmp;
			}
			C[0] = 0;

			// get file for BWT
			array_sequential_write BWT(bwt_filename, ((n-1)/8+2)*8, 0);
			for(size_t i=0, bits = n*8; i<8; ++i)
			{
				BWT.set(i, bits);
				bits >>= 8;
			}

			// initiate bucket boundaries for correct BWT write operations
			seqtext.begin_backward_read();
			vector<unsigned long long> bucket_front(realsigma);
			vector<unsigned long long> bucket_end(realsigma);
			for(unsigned long long i=0; i<realsigma; ++i)
			{
				bucket_front[i] = C[i];
				bucket_end[i] = C[i+1]-1;
			}

			BWT.set(0+8, number_to_char[seqtext.at(n-2)]);
			unsigned char c = seqtext.at(n-2);
			size_t i = n-3;
			unsigned char b = seqtext.at(i);

			//move right
			while(b>=c)
			{
				BWT.set(bucket_front[c]+8, number_to_char[b]);
				++bucket_front[c];
				c = b;
				--i;
				b = seqtext.at(i);
			}

			//move left
			BWT.start_backward_read_and_write();
			while(i+1>0)
			{
				BWT.set(bucket_end[c]+8, number_to_char[b]);
				--bucket_end[c];
				c = b;
				--i;
				b = seqtext.at(i);
			}
			sdsl::remove(text_trans_filename);
			cout << "recursion level=0" << "  n=" << n << "  sigma=" << sigma << "  lms_short_bound=" << lms_short_bound << "  lms_long_bound=" << lms_long_bound << endl;
			return;
		}

		// usually the input consists of more than one lms_string
		// find all the other lms strings
		for(size_t i=startpos; i+1>0; --i)
		{
			++C[seqtext.at(i)];
			if(smaller)
			{
				if(seqtext.at(i)>seqtext.at(i+1))
				{
					if(last-i>lms_short_bound) //long
					{
						if(last-i>longestlong) longestlong = last-i;
						lms_value = 0;
						for(size_t j=i+1; j<last+1 and j<i+lms_long_bound+1; ++j)
						{
							lms_value = lms_value*sigma+(seqtext.at(j)-1);
						}
						if(last-i<lms_long_bound)
						{
							for(size_t j=0; j<lms_long_bound-(last-i); ++j)
							{
								lms_value = lms_value*sigma+(sigma-1);
							}
						}
						else if(last-i>lms_long_bound)
						{
							for(size_t j=i+lms_long_bound+1; j<last+1; ++j)
							{
								text_longs.push_back(seqtext.at(j));
							}
						}
						lms_long.push_back(triple(lms_value,text_longs.size()-(last-i)+lms_long_bound,text_longs.size(),nrs));
				}
				else //short
				{
					lms_value = 0;
					for(size_t j=i+1; j<last+1; j++)
					{
						lms_value = lms_value*sigma+(seqtext.at(j)-1);
					}
					for(size_t j=0; j<lms_short_bound-(last-i); ++j)
					{
						lms_value = lms_value*sigma+(sigma-1);
					}
					if(lms_short[lms_value]==0) ++nrs_s;
					lms_short[lms_value] = 1;
					}
					nrs++;
					smaller = false;
					last = i+1;
				}
			}
			else // last was larger, find smaller
			{
				if(seqtext.at(i)<seqtext.at(i+1))
				{
					smaller = true;
				}
			}
		}
		seqtext.finish_this_read();

		// calculate C-Array
		for(unsigned long long i=1, tmp=0, tmp2=0; i<C.size(); ++i)
		{
			tmp = C[i];
			C[i] = tmp2+C[i-1];
			tmp2 = tmp;
		}
		C[0] = 0;

		// Step 1b: sort longs
		cout << "Step 1b - sort long lms strings" << endl;

		unsigned char * datapointer = &(text_longs[0]);
		text_longs_comp text_lo(datapointer, text_longs.size());
		sort(lms_long.begin(), lms_long.end(), text_lo);


		// Step 2:
		//calculate lexicographical names,
		//write longs in translation table name_to_pos
		//prepare write of longs into reduced text red_text
		cout << "Step 2a - calculate lexicographical names, write longs into translation table and prepare write of reduced text" << endl;

		rank_support_v<> short_rank_support;                  // Rank support for bit_vector lms_short
		util::init_support(short_rank_support, &lms_short);   // Create rank support
		size_t names_max = lms_long.size()+short_rank_support.rank(lms_short.size());
		bit_vector names(names_max+1, false);                 // bit_vector indicates if i is name of a short
		int_vector<64> name_to_pos(nrs_s+lms_long.size(),0);  // translation table
		int_vector<1> long_for_redtext(nrs,0);                // remember longes to write in red_text (step 3)
		size_t resize_size = 100+longestlong;
		size_t name_count = 0;                                // next name to give to a lms-string (0 is for $...)
		size_t long_idx = 0;                                  // index of the next long lms-string that will be named
		size_t red_idx = 0;

		//calculate short_value of first long
		size_t it_value = lms_long[long_idx].value;
		for(size_t i=lms_short_bound; i<lms_long_bound; ++i)
		{
			it_value /= sigma;
		}

		// go through bit_vector lms_short and the sorted longs in parallel
		for(size_t i=0; i<lms_short.size(); ++i)
		{
			//go through long lms-strings of (translated short) lms_value i
			while(it_value==i)
			{
				++name_count;
				if(lms_long[long_idx].pos<lms_long[long_idx].pos_ex)
				{
					red_idx = red_idx + (lms_long[long_idx].pos_ex-lms_long[long_idx].pos) + lms_long_bound;
				}
				else
				{
					red_idx = red_idx + (lms_long[long_idx].pos_ex-(lms_long[long_idx].pos-lms_long_bound));
				}
				name_to_pos[name_count] = red_idx;
				++red_idx;
				long_for_redtext[lms_long[long_idx].nr] = 1;
				if(lms_long[long_idx].nr==1) //...$
				{
					lastlms_name = name_count;
				}
				++long_idx;
				while(long_idx<lms_long.size() and (text_lo.equal(lms_long[long_idx-1],lms_long[long_idx])))
				{
					// identical longs receive the same name
					lms_long[long_idx-1].value = name_count;
					++long_idx;
				}
				lms_long[long_idx-1].value = name_count;
				if(long_idx<lms_long.size())
				{
					//calculate short value of next lms_long
					it_value = lms_long[long_idx].value/pow(sigma,lms_long_bound-lms_short_bound);
				}
				else
				{
					//no more longs
					break;
				}
			}
			//short lms-string
			if(lms_short[i])
			{
				++name_count;
				names[name_count] = 1; // mark short name
			}
		}

		text_longs.resize(0);

		// Step 2b: write longs in text_rek (text for recursion)
		cout << "Step 2b - write longs in text for recursion" << endl;

		sigma_rek = name_count+1;                           //sigma of next recursion
		text_rek.width(bits::hi(sigma_rek)+1);
		text_rek.resize(nrs);
		for(size_t i=0; i<lms_long.size();++i)
		{
			text_rek[nrs-1-lms_long[i].nr]=lms_long[i].value;
		}

		// Step 3: generate new text
		cout << "Step 3  - write shorts in text for recursion and longs into reduced text" << endl;

		select_support_mcl<> name_select_support;           // select_support for bit_vector
		util::init_support(name_select_support, &names);    // Create select_support
		int_vector<8> red_text(red_idx+1, 1);               // reduced text contains all the serent lms strings, for use in step 5-7
		size_t nrs_new = 1;
		size_t lms_pos = nrs-1;                              // index into text_rek
		startpos = n-2;
		last = n-1;
		smaller = false;
		seqtext.begin_backward_read();
		//$
		text_rek[lms_pos] = 0;
		--lms_pos;
		//...$
		for(size_t i=startpos; i+1>0; --i)
		{
			if(smaller)
			{
				if(seqtext.at(i)>seqtext.at(i+1))
				{
					size_t lopt = name_to_pos[text_rek[lms_pos]]; //last position of this lms string in red_text
					lopt = lopt-(last-i);
					red_text[lopt] = sigma-1; //delim
					++lopt;
					for(size_t j=i+1; j<last; j++)
					{
						red_text[lopt] = seqtext.at(j)-1;
						++lopt;
					}
					red_text[lopt] = 0;  //this is $
					--lms_pos;
					++nrs_new;
					smaller = false;
					last = i+1;
					startpos = i-1;
					break;
				}
			}
			else
			{
				if(seqtext.at(i)<seqtext.at(i+1))
				{
					smaller = true;
				}
			}
		}

		// all the other lms_strings
		for(size_t i=startpos; i+1>0; --i)
		{
			if(smaller)
			{
				if(seqtext.at(i)>seqtext.at(i+1))
				{
					if(last-i>lms_short_bound) //longs
					{
						if(long_for_redtext[nrs_new])
						{
							size_t lopt = name_to_pos[text_rek[lms_pos]]-(last-i); //first position of this lms string in red_text
							red_text[lopt] = sigma-1; //delim
							++lopt;
							for(size_t j=i+1; j<last+1; j++)
							{
								red_text[lopt] = seqtext.at(j)-1;
								++lopt;
							}
						}
						--lms_pos;
					}
					else //shorts
					{
						lms_value = 0;
						for(size_t j=i+1; j<last+1; j++)
						{
							lms_value = lms_value*sigma+(seqtext.at(j)-1);
						}
						for(size_t j=0; j<lms_short_bound-(last-i); ++j)
						{
							lms_value = lms_value*sigma+(sigma-1);
						}
						// find the name of this short and write it to the text for recursion
						text_rek[lms_pos] = name_select_support.select(short_rank_support.rank(lms_value)+1);
						//translation table and reduced text:
						if(name_to_pos[text_rek[lms_pos]] == 0)
						{
							while(red_text.size()<red_idx+longestlong+1)
							{
								red_text.resize(red_text.size()+resize_size);
							}
							red_text[red_idx] = sigma-1;  //delimiter
							++red_idx;
							for(size_t j=i+1; j<last+1; j++)  //incl delimiter
							{
								red_text[red_idx] = seqtext.at(j)-1;
								++red_idx;
							}
							name_to_pos[text_rek[lms_pos]] = red_idx-1;
						}
						--lms_pos;
					}
					++nrs_new;
					smaller = false;
					last = i+1;
				}
			}
			else
			{
				if(seqtext.at(i)<seqtext.at(i+1))
				{
					smaller = true;
				}
			}
		}

		//lms_string $...
		while(red_text.size()<red_idx+last+4)
		{
			red_text.resize(red_idx+last+4);
		}
		//remember: 0 before delimiter has to be substituted by $ when translating the bwt from recursion!!!
		red_text[red_idx] = sigma-1;  //delimiter
		++red_idx;
		red_text[red_idx] = 0; //substitutes $
		dollar_pos = red_idx;
		++red_idx;
		for(size_t i=0; i<last+1; ++i)
		{
			red_text[red_idx] = seqtext.at(i)-1;
			++red_idx;
		}
		name_to_pos[0] = red_idx-1;

		//remove excess space
		util::clear(lms_long);
		red_text.resize(red_idx);
		name_to_pos.resize(sigma_rek+1);

		//store C-Array, name_to_pos, redtext
		red_text_size = red_text.size();
		name_to_pos_size = name_to_pos.size();
		store_to_file(name_to_pos, filename_translation.c_str());
		store_to_file(red_text, filename_redtext.c_str());
		sdsl::remove(text_trans_filename);
	}


	//Step 4: Recursion returns bwt of text_rek
	cout << "Step 4  - recursion" << endl;

	int_vector<> bwt_rek;
	bwt_rek.width(text_rek.width());
	std::string text_filename_rek = "text_rek_0";
	store_to_file(text_rek, text_filename_rek);
	if(sigma_rek < (1ULL<<8))
	{
		if(text_rek.size() < (1ULL<<32))
		{
			_construct_bwt_from_text<uint8_t, uint32_t, int_vector<> >(text_rek, bwt_rek, sigma_rek, 1);
		}
		else
		{
			_construct_bwt_from_text<uint8_t, uint64_t, int_vector<> >(text_rek, bwt_rek, sigma_rek, 1);
		}
	}
	else if(sigma_rek < (1ULL<<16))
	{
		if(text_rek.size() < (1ULL<<32))
		{
			_construct_bwt_from_text<uint16_t, uint32_t, int_vector<> >(text_rek, bwt_rek, sigma_rek, 1);
		}
		else
		{
			_construct_bwt_from_text<uint16_t, uint64_t, int_vector<> >(text_rek, bwt_rek, sigma_rek, 1);
		}
	}
	else if(sigma_rek < (1ULL<<32))
	{
		if(text_rek.size() < (1ULL<<32))
		{
			_construct_bwt_from_text<uint32_t, uint32_t, int_vector<> >(text_rek, bwt_rek, sigma_rek, 1);
		}
		else
		{
			_construct_bwt_from_text<uint32_t, uint64_t, int_vector<> >(text_rek, bwt_rek, sigma_rek, 1);
		}
	}
	else
	{
		_construct_bwt_from_text<uint64_t, uint64_t, int_vector<> >(text_rek, bwt_rek, sigma_rek, 1);
	}
	remove(text_filename_rek.c_str());

	//Step 5: Initiate datastructures needed for creation of BWT
	cout << "Step 5  - init datastructures for induced sorting" << endl;

	//load C-Array, name_to_pos, red_text
	int_vector<64> name_to_pos(name_to_pos_size,0);
	int_vector<8> red_text(red_text_size);
	load_from_file(name_to_pos, filename_translation.c_str());
	remove(filename_translation.c_str());
	load_from_file(red_text, filename_redtext.c_str());
	remove(filename_redtext.c_str());
	uint8_t int_width = bits::hi(red_text.size()-1)+1;
	size_t resize_block = 1000;
	if(resize_block < n/(8*100*sigma))
	{
		resize_block = n/(8*100*sigma);
	}
	vector< int_vector<> > L_QUEUE(realsigma, int_vector<>(0, 0, int_width));

	vector<unsigned long long> LMS_begin(realsigma,0); //points to first element in LMS_QUEUE
	vector<unsigned long long> LMS_ex(realsigma,0); //points to space behind last element in LMS_QUEUE
	vector<unsigned long long> L_begin(realsigma,0); //points to first element in L_QUEUE
	vector<unsigned long long> L_ex(realsigma,0); //points to space behind last element in L_QUEUE
	vector<unsigned long long> S_begin(realsigma,0); //points to first element in S_QUEUE
	vector<unsigned long long> S_ex(realsigma,0); //points to space behind last element in S_QUEUE
	vector<unsigned long long> bucket_front(realsigma);
	vector<unsigned long long> bucket_end(realsigma);
	vector< cached_external_array<uint64_t> > LMS_QUEUE_c(realsigma, cached_external_array<uint64_t>());

	for(size_t i=0; i<realsigma; ++i)
	{
		LMS_QUEUE_c[i].init("lms_queue_"+i, 1024*1024/8, true, true);
	}

	for(unsigned long long i=0; i<realsigma; ++i)
	{
		bucket_front[i] = C[i];
		bucket_end[i] = C[i+1]-1;
	}
	for(unsigned long long i=0; i<bwt_rek.size(); ++i)
	{
		if(bwt_rek[i]==lastlms_name) //last lms-string ends with $
		{
			unsigned long long pos = name_to_pos[bwt_rek[i]];
			LMS_QUEUE_c[0].write(0,pos-1);
			LMS_ex[0]=1;
		}
		else
		{
			unsigned long long pos = name_to_pos[bwt_rek[i]];
			unsigned char c = red_text[pos]+1;
			LMS_QUEUE_c[c].write(LMS_ex[c],pos-1);
			++LMS_ex[c];
		}
	}

	util::clear(name_to_pos);
	util::clear(bwt_rek);

	array_sequential_write BWT(bwt_filename, ((n-1)/8+2)*8, 0);
	for(size_t i=0, bits = n*8; i<8; ++i)
	{
		BWT.set(i, bits);
		bits >>= 8;
	}

	cached_external_array<uint64_t> L_STACK_c("l_stack_0", 1024*1024, false, true);
	unsigned long long L_stack_top = 0;  // next empty stack position
	cached_external_array<uint8_t> char_stack_c("char_stack_0", 1024*1024, false, true);
	unsigned long long char_stack_top = 0;

	//Step 6: go through array left to right
	cout << "Step 6  - go through array left to right" << endl;

	for(unsigned long c=0; c<realsigma; ++c)
	{
		while(L_begin[c]!=L_ex[c])
		{
			size_t queue_end = L_ex[c];
			L_ex[c] = 0;
			while(L_begin[c]!=queue_end)
			{
				unsigned long long pos = L_QUEUE[c][L_begin[c]];
				++L_begin[c];
				unsigned char b = red_text[pos]+1;
				if(b<c)
				{
					L_STACK_c.write(L_stack_top++, pos);
					//catch ba.....$
					if(pos==dollar_pos)
					{
						b = 0;
					}
				}
				else
				{
					if(L_ex[b]>=L_QUEUE[b].size())
					{
						L_QUEUE[b].resize(L_QUEUE[b].size()+resize_block);
					}
					L_QUEUE[b][L_ex[b]] = pos-1;
					++L_ex[b];
				}
				BWT.set(bucket_front[c]+8, number_to_char[b]);
				++bucket_front[c];
			}
			L_begin[c] = 0;
			L_QUEUE[c].resize(L_ex[c]);
		}
		L_QUEUE[c].resize(0);
		while(LMS_begin[c]!=LMS_ex[c])
		{
			unsigned long long pos = LMS_QUEUE_c[c].read(LMS_begin[c]);
			++LMS_begin[c];
			unsigned char b = red_text[pos]+1;
			char_stack_c.write(char_stack_top++,b);
			if(L_ex[b]>=L_QUEUE[b].size())
			{
				L_QUEUE[b].resize(L_QUEUE[b].size()+resize_block);
			}
			L_QUEUE[b][L_ex[b]] = pos-1;
			++L_ex[b];
		}
	}
	BWT.start_backward_read_and_write();

	vector< int_vector<> > S_QUEUE(realsigma, int_vector<>(0, 0, int_width));
	S_QUEUE.swap(L_QUEUE);

	//Step 7: go through array right to left
	cout << "Step 7  - go through array right to left" << endl;
	for(unsigned long c=realsigma-1; c+1>0; --c)
	{
		while(S_begin[c]!=S_ex[c])
		{
			size_t queue_end = S_ex[c];
			S_ex[c] = 0;
			while(S_begin[c]!=queue_end)
			{
				unsigned long long pos = S_QUEUE[c][S_begin[c]];
				++S_begin[c];
				unsigned char b;
				if(red_text[pos]>red_text[pos+1] or pos+1==dollar_pos) //found delimiter
				{
					b = char_stack_c.read(--char_stack_top);
				}
				else
				{
					b = red_text[pos]+1;
					if(pos==dollar_pos)
					{
						b = 0;
					}
					if(S_ex[b]>=S_QUEUE[b].size())
					{
						S_QUEUE[b].resize(S_QUEUE[b].size()+resize_block);
					}
					S_QUEUE[b][S_ex[b]] = pos-1;
					++S_ex[b];
				}
				BWT.set(bucket_end[c]+8, number_to_char[b]);
				--bucket_end[c];
			}
			S_begin[c] = 0;
			S_QUEUE[c].resize(S_ex[c]);
		}
		S_QUEUE[c].resize(0);
		while(L_stack_top!=0)
		{
			unsigned long long pos = L_STACK_c.read(L_stack_top-1);
			if((unsigned long)red_text[pos+1]+1!=c)
			{
				break;
			}
			unsigned char b = red_text[pos]+1;
			--L_stack_top;
			if(pos==dollar_pos)
			{
				b = 0;
			}
			if(S_ex[b]>=S_QUEUE[b].size())
			{
				S_QUEUE[b].resize(S_QUEUE[b].size()+resize_block);
			}
			S_QUEUE[b][S_ex[b]] = pos-1;
			++S_ex[b];
		}
	}
	util::clear(S_QUEUE);
	cout << "recursion level=0" << "  n=" << n << "  sigma=" << sigma << "  lms_short_bound=" << lms_short_bound << "  lms_long_bound=" << lms_long_bound << endl;
}



//! Constructs the Burrows and Wheeler Transform (BWT) from text over byte-alphabet.
/*! The algorithm constructs the BWT and stores it to disk.
 *  \param config	Reference to cache configuration
 *  \par Space complexity
 *		\f Usually less than $1.5n \f$ bytes
 *  \pre Text exist in the cache. Keys:
 *         * constants::KEY_TEXT
 *  \post BWT exist in the cache. Key
 *         * constants::KEY_BWT
 */
void construct_bwt_from_text(cache_config& config)
{
	string text_filename = cache_file_name(constants::KEY_TEXT, config);
	string bwt_filename = cache_file_name(constants::KEY_BWT, config);
	cout << "bwt_filename = " << bwt_filename << endl;
	int_vector<8> text;
	load_from_file(text, text_filename);
	size_t n = text.size();

	if(n==2)
	{
		//if text is c$ write bwt c$
		store_to_cache(text, constants::KEY_BWT, config);
		return;
	}

	// Find Alphabet Size Sigma
	vector<bool> used_chars(256,0);
	for (size_t i=0; i < n; ++i) {
		used_chars[text[i]]=1;
	}
	size_t sigma = 0;
	for(size_t i=0; i<used_chars.size(); ++i)
	{
		if(used_chars[i])
		{
			++sigma;
		}
	}
	--sigma;

	if(sigma > 8)
	{
		// call construct_bwt_larger_alphabet
		cout << "call construct_bwt_larger_alphabet" << endl;
		int_vector<8> bwt;
		sigma = 256;
		if(text.size() < (1ULL<<32))
		{
			_construct_bwt_from_text<uint8_t, uint32_t, int_vector<8> >(text, bwt, sigma, 0);
		}
		else
		{
			_construct_bwt_from_text<uint8_t, uint64_t, int_vector<8> >(text, bwt, sigma, 0);
		}
		store_to_cache(bwt, constants::KEY_BWT, config);
	}
	else
	{
		// call construct_bwt_smaller_alphabet
		cout << "call construct_bwt_smaller_alphabet" << endl;
		unsigned long long powof2sigma = 1;
		size_t char_bits = 0;
		while(powof2sigma < sigma)
		{
			powof2sigma <<= 1;
			++char_bits;
		}
		size_t lms_long_bound = 0;
		for(size_t max_value=-1; max_value>0; lms_long_bound++)
		{
			max_value /= powof2sigma;
		}
		size_t lms_short_bound = 8;
		construct_bwt_smaller_alphabet(bwt_filename, text, used_chars, lms_short_bound, lms_long_bound);
		register_cache_file(constants::KEY_BWT, config);
	}
}

int main(int argc, char** argv)
{

	// Check Parameters
	if(argc!=3)
	{
        cout << "Usage: " << argv[0] << " input_file file_name" << endl;
		cout << endl;
		cout << "Outputfile: text_file_name.sdsl and bwt_file_name.sdsl" << endl;
		return 1;
	}

	// Get parameters
	string file = argv[1];
	string dir = ".";
	string id = argv[2];
	cache_config config(false, dir, id);

	{// (1) convert input_file to sdsl-format
		int_vector<8> text;
		load_vector_from_file(text, file, 1);
		if ( contains_no_zero_symbol(text, file) ){
			append_zero_symbol(text);
			store_to_cache(text, constants::KEY_TEXT, config);
		}
		register_cache_file(constants::KEY_TEXT, config);
	}

	{//  (2) construct BWT
		construct_bwt_from_text(config);
	}

	return 0;
}
